/*
 * Copyright (c) 2020 Tilman Sauerbeck (tilman at code-monkey de)
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

use gpio;
use spi;
use storage::{Storage, Error};

const SECTOR_SIZE: usize = 4096;
const PAGE_SIZE:   usize = 256;

pub struct Mx25l {
    cs_gpio: u32,
    cs_gpio_pin: u32,
}

enum Command {
    PP   = 0x02,
    READ = 0x03,
    RDSR = 0x05,
    WREN = 0x06,
    SE   = 0x20,
    CE   = 0x60,
    RDID = 0x9f,
}

const SR_WIP: u8 = 1 << 0;

impl Mx25l {
    pub fn new(cs_gpio: u32, cs_gpio_pin: u32) -> Mx25l {
        Mx25l {
            cs_gpio: cs_gpio,
            cs_gpio_pin: cs_gpio_pin,
        }
    }

    pub fn read_status(&self) -> u8 {
        self.with_selected(|| {
            spi::tx8(spi::SPI0, Command::RDSR as u8);

            spi::tx8(spi::SPI0, 0xff)
        })
    }

    pub fn read_id(&self) -> (u8, u16) {
        self.with_selected(|| {
            spi::tx8(spi::SPI0, Command::RDID as u8);

            let manufacturer_id = spi::tx8(spi::SPI0, 0xff);
            let device_id0 = spi::tx8(spi::SPI0, 0xff) as u16;
            let device_id1 = spi::tx8(spi::SPI0, 0xff) as u16;

            (manufacturer_id, device_id0 | (device_id1 << 8))
        })
    }

    pub fn erase_all(&self) {
        self.write_enable();

        self.with_selected(|| {
            spi::tx8(spi::SPI0, Command::CE as u8);
        });

        while self.write_in_progress() {
        }
    }

    pub fn erase_sector(&self, address: usize) -> Result<(), Error> {
        if (address & (SECTOR_SIZE - 1)) != 0 {
            return Err(Error::UnalignedAddress);
        }

        self.write_enable();

        self.with_selected(|| {
            spi::tx8(spi::SPI0, Command::SE as u8);

            spi::tx8(spi::SPI0, (address >> 16) as u8);
            spi::tx8(spi::SPI0, (address >>  8) as u8);
            spi::tx8(spi::SPI0, (address >>  0) as u8);
        });

        while self.write_in_progress() {
        }

        Ok(())
    }

    pub fn program_page(&self, address: usize, buffer: &[u8; PAGE_SIZE])
                        -> Result<(), Error> {
        if (address & (PAGE_SIZE - 1)) != 0 {
            return Err(Error::UnalignedAddress);
        }

        if buffer.iter().all(|&b| b == 0xff) {
            return Ok(());
        }

        self.with_selected(|| {
            spi::tx8(spi::SPI0, Command::PP as u8);

            spi::tx8(spi::SPI0, (address >> 16) as u8);
            spi::tx8(spi::SPI0, (address >>  8) as u8);
            spi::tx8(spi::SPI0, (address >>  0) as u8);

            for &b in buffer.iter() {
                spi::tx8(spi::SPI0, b);
            }
        });

        while self.write_in_progress() {
        }

        Ok(())
    }

    fn write_enable(&self) {
        self.with_selected(|| {
            spi::tx8(spi::SPI0, Command::WREN as u8);
        });
    }

    fn write_in_progress(&self) -> bool {
        (self.read_status() & SR_WIP) != 0
    }

    fn with_selected<F, T>(&self, func: F) -> T
        where F: FnOnce() -> T
    {
        gpio::clear(self.cs_gpio, self.cs_gpio_pin);

        let r = func();

        gpio::set(self.cs_gpio, self.cs_gpio_pin);

        r
    }
}

impl Storage for Mx25l {
    fn size(&self) -> usize {
        2 << 20
    }

    fn read(&self, address: usize, buffer: &mut [u8]) {
        self.with_selected(|| {
            spi::tx8(spi::SPI0, Command::READ as u8);

            spi::tx8(spi::SPI0, (address >> 16) as u8);
            spi::tx8(spi::SPI0, (address >>  8) as u8);
            spi::tx8(spi::SPI0, (address >>  0) as u8);

            for i in 0..buffer.len() {
                buffer[i] = spi::tx8(spi::SPI0, 0xff);
            }
        })
    }

    fn write(&mut self, address: usize, buffer: &[u8; SECTOR_SIZE])
             -> Result<(), Error> {
        if let Err(e) = self.erase_sector(address) {
            return Err(e);
        }

        for (i, page_bytes) in buffer.chunks(PAGE_SIZE).enumerate() {
            let page_address = address + (i * PAGE_SIZE);

            // XXX: Inefficient.
            let mut ba = [0xff; PAGE_SIZE];
            ba.copy_from_slice(&page_bytes);

            if let Err(e) = self.program_page(page_address, &ba) {
                return Err(e);
            }
        }

        Ok(())
    }

    fn clear(&mut self) {
        self.erase_all();
    }
}
